----------------------------------------------------------------------------
-- Interleaved Multiplier (LSB_first_squarer_V2.vhd)
--
-- LSB first squaring
-- 
-- Computes the polynomial squaring mod f in GF(2**m)
-- Implements a sequential cincuit

-- 
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package LSB_first_squarer_package is
  constant M: integer := 8;
  constant F: std_logic_vector(M-1 downto 0):= "00011011"; --for M=8 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"001B"; --for M=16 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0101001B"; --for M=32 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0000000000000000010100000101001B"; --for M=128 bits
  --constant F: std_logic_vector(M-1 downto 0):= "000"&x"00000000000000000000000000000000000000C9"; --for M=163
  --constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
  function product_alphaF return std_logic_vector;
end LSB_first_squarer_package;

package body LSB_first_squarer_package is
  function product_alphaF return std_logic_vector is
    variable Faux: std_logic_vector(M-1 downto 0);
  begin
    Faux(0) := F(M-1) and F(0);
    for i in 1 to M-1 loop 
      Faux(i) := (F(i-1) xor (F(M-1) and F(i))); 
    end loop;
    return Faux;
  end function;
end LSB_first_squarer_package;

-----------------------------------
-- LSB-first squarer cell
-----------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.LSB_first_squarer_package.all;

entity LSB_first_squarer_cell is
port (
  b: in std_logic_vector(M-1 downto 0);
  w: in std_logic_vector(M-1 downto 0);
  a_k: in std_logic;
  new_b: out std_logic_vector(M-1 downto 0);
  new_w: out std_logic_vector(M-1 downto 0)
);
end LSB_first_squarer_cell;

architecture rtl of LSB_first_squarer_cell is
  constant Faux: std_logic_vector(M-1 downto 0) := product_alphaF;
begin

  new_b_calc: for i in 2 to M-1 generate
      new_b(i) <= B(i-2) xor (B(m-1) and Faux(i)) xor (B(m-2) and F(i));
  end generate;  
  new_b(1) <= (B(m-1) and Faux(1)) xor (B(m-2) and F(1));
  new_b(0) <= (B(m-1) and Faux(0)) xor (B(m-2) and F(0));

  new_w_calc: for i in 0 to M-1 generate
    new_w(i) <= w(i) xor (b(i) and a_k);
  end generate;  

end rtl;

-----------------------------------
-- LSB_first_squarer
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.LSB_first_squarer_package.all;

entity LSB_first_squarer is
port (
  A: in std_logic_vector (M-1 downto 0);
  clk, reset, start: in std_logic; 
  Z: out std_logic_vector (M-1 downto 0);
  done: out std_logic
);
end LSB_first_squarer;

architecture rtl of LSB_first_squarer is

  component LSB_first_squarer_cell is
  port (
    b: in std_logic_vector(M-1 downto 0);
    w: in std_logic_vector(M-1 downto 0);
    a_k: in std_logic;
    new_b: out std_logic_vector(M-1 downto 0);
    new_w: out std_logic_vector(M-1 downto 0)
  );
  end component;

  signal inic, shift_r, ce_c: std_logic;
  signal count: natural range 0 to (M+1)/2;
  type states is range 0 to 3;
  signal current_state: states;
  signal b, w: std_logic_vector(M-1 downto 0);
  signal aa: std_logic_vector(M/2-1 downto 0);
  signal new_b, new_w: std_logic_vector(M-1 downto 0);
begin
  
   basicCell: LSB_first_squarer_cell port map (
                            b => b, w => w, a_k => aa(0), 
                            new_b => new_b, new_w => new_w );

  register_A: process(reset, clk)
  begin
    if reset = '1' then aa <= (others => '0');
    elsif clk'event and clk = '1' then
      if inic = '1' then
         aa <= a(M-1 downto (M+1)/2);
      else
         aa <= '0' & aa(M/2-1 downto 1);
      end if;
    end if;
  end process register_A;

  register_b: process(reset, clk)
  begin
    if reset = '1' then b <= (others => '0');
    elsif clk'event and clk = '1' then
      if inic = '1' then 
        if M mod 2 = 0 then b <= F;
        else b <= product_alphaF; end if;
      elsif ce_c = '1' then 
        b <= new_b;
      end if;
    end if;
  end process register_b;
  
  register_w: process(reset, inic, clk)
  begin
    if reset = '1' then w <= (others => '0');
    elsif clk'event and clk = '1' then
      if inic = '1' then
        w(0) <= A(0);
        square: for i in 1 to (M-1)/2 loop
          w(2*i-1) <= '0'; w(2*i) <= A(i);
        end loop;
        if M mod 2 = 0 then w(M-1) <= '0'; end if;
      elsif ce_c = '1' then 
        w <= new_w;
      end if;
    end if;
  end process register_w;
  
  z <= w;

  counter: process(reset, clk)
  begin
    if reset = '1' then count <= 0;
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        count <= 0;
      elsif shift_r = '1' then
        count <= count+1; 
    end if;
    end if;
  end process counter;

  control_unit: process(clk, reset, current_state)
  begin
    case current_state is
      when 0 to 1 => inic <= '0'; shift_r <= '0'; done <= '1'; ce_c <= '0';
      when 2 => inic <= '1'; shift_r <= '0'; done <= '0'; ce_c <= '0';
      when 3 => inic <= '0'; shift_r <= '1'; done <= '0'; ce_c <= '1';
    end case;

    if reset = '1' then current_state <= 0;
    elsif clk'event and clk = '1' then
      case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if count = (M+1)/2-1 then current_state <= 0; end if;
      end case;
    end if;
  end process control_unit;

end rtl;
