----------------------------------------------------------------------------
--  Modified Almost Inverse Algorithm for inversion (MAIA_inversion.vhd)
--
-- Computes the 1/x mod f in GF(2**m)
-- Implements a sequential cincuit 
-- 
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package maia_inversion_package is
  constant M: integer := 8;
  constant logM: integer := 3;
  constant F: std_logic_vector(M-1 downto 0):= "00011011"; --for M=8 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"001B"; --for M=16 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0101001B"; --for M=32 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"010100000101001B"; --for M=64 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0000000000000000010100000101001B"; --for M=128 bits
  --constant F: std_logic_vector(M-1 downto 0):= "000"&x"00000000000000000000000000000000000000C9"; --for M=163
  --constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
end maia_inversion_package;


-----------------------------------
-- MAIIA_inversion data_path
-----------------------------------
library ieee; 
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use work.maia_inversion_package.all;

entity maia_inversion_data_path is
port (
  b, c: in std_logic_vector(M downto 0);
  u, v: in std_logic_vector(M downto 0);
  degree_u, degree_v: in STD_LOGIC_VECTOR (logM downto 0);
  new_b, new_c: out std_logic_vector(M downto 0);
  new_u, new_v: out std_logic_vector(M downto 0);
  new_degree_u, new_degree_v: out STD_LOGIC_VECTOR (logM downto 0)
);
end maia_inversion_data_path;

architecture rtl of maia_inversion_data_path is
  constant zero: std_logic_vector(logM downto 0):= (others => '0');

begin

  process(b,c,u,v,degree_u,degree_v)
    variable n_U, n_V, n_A, n_B, auxm: std_logic_vector(M downto 0);
    variable d: natural;
  begin
      if U(0) = '0' then
         new_u <= '0' & u(M downto 1);
         new_degree_u <= degree_u - 1;
         if B(0) = '0' then
            new_b <= '0' & b(M downto 1);
         else
            new_b <= '0' & (b(M downto 1) xor ('1' & F(M-1 downto 1)) );
         end if;
      else
        new_U <= U xor V;
        new_B <= B xor C;
        d := 0; --degree calculation
        for i in 0 to m loop
           if (U(i) xor V(i)) = '1' then d := i; end if;
        end loop;
        new_degree_u <= conv_std_logic_vector(d,logM+1);
      end if;
  end process;

   
  process(b,c,u,v,degree_u, degree_v)
  begin
    if degree_u < degree_v then 
       new_v <= u;
       new_c <= b;
       new_degree_v <= degree_u;
    else
       new_v <= v;
       new_c <= c;
       new_degree_v <= degree_v;
    end if;
  end process;

end rtl;

-----------------------------------
--  (MAIA_inversion)
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.maia_inversion_package.all;

entity maia_inversion is
port (
  A: in std_logic_vector (M-1 downto 0);
  clk, reset, start: in std_logic; 
  Z: out std_logic_vector (M-1 downto 0);
  done: out std_logic
);
end maia_inversion;

architecture rtl of maia_inversion is

  COMPONENT maia_inversion_data_path
  PORT(
    b, c: in std_logic_vector(M downto 0);
    u, v: in std_logic_vector(M downto 0);
    degree_u, degree_v: in STD_LOGIC_VECTOR (logM downto 0);
    new_b, new_c: out std_logic_vector(M downto 0);
    new_u, new_v: out std_logic_vector(M downto 0);
    new_degree_u, new_degree_v: out STD_LOGIC_VECTOR (logM downto 0) );
  END COMPONENT;

  signal count: natural range 0 to 2*M;
  type states is range 0 to 3;
  signal current_state: states;
  
  
  signal first_step, ce_vc, ce_ub, end_inv: std_logic; 
  signal b, c, new_b, new_c : std_logic_vector(M downto 0);
  signal u, v, new_u, new_v: std_logic_vector(M downto 0);
  signal degree_u, new_degree_u, degree_v, new_degree_v, degree_a: std_logic_vector(logM downto 0);
  constant ONE: std_logic_vector(M downto 0):= (0 => '1', others => '0');
begin

  data_path_block: maia_inversion_data_path PORT MAP(
      b => b, c => c,
      u => u, v => v, degree_u => degree_u, degree_v => degree_v,
      new_b => new_b, new_c => new_c,
      new_u => new_u, new_v => new_v, 
      new_degree_u => new_degree_u, new_degree_v => new_degree_v);
  
  z <= c(M-1 downto 0);

  degr_a: process(clk, reset)
    variable d: natural;
  begin
    d := 0; 
    for i in 0 to m-1 loop
      if (A(i)) = '1' then d := i; end if;
    end loop;
    degree_a <= conv_std_logic_vector(d,logM+1);
  end process;

  reg: process(clk, reset)
    variable d: natural;
  begin
    if reset = '1' then 
       u <= (others => '0'); v <= (others => '0');
       b <= (others => '0'); c <= (others => '0');
       degree_v <= (others => '0'); degree_u <= (others => '0');
    elsif clk'event and clk = '1' then
      if first_step = '1' then 
         v <= ('1' & F); c <= (others => '0');
         degree_v <= conv_std_logic_vector(M, logM+1);
      elsif ce_vc = '1' then
        v <= new_v; c <= new_c;
        degree_v <= new_degree_v;
      end if;
      if first_step = '1' then
        u <= '0' & A; b <= (0 => '1', others => '0');
        degree_u <= degree_a; 
      elsif ce_ub = '1' then
        u <= new_u; b <= new_b;
        degree_u <= new_degree_u;
      end if;
    end if;
  end process;

  end_inv <= '1' when U = ONE else '0';

  control_unit: process(clk, reset, current_state, count)
  begin
    case current_state is
      when 0 to 1 => first_step <= '0'; done <= '1'; ce_vc <= '0'; ce_ub <= '0';
      when 2 => first_step <= '1'; done <= '0'; ce_vc <= '0'; ce_ub <= '0';
      when 3 => first_step <= '0'; done <= '0'; ce_vc <= U(0); ce_ub <= '1';
    end case;
  
    if reset = '1' then current_state <= 0;
    elsif clk'event and clk = '1' then
      case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3;
      when 3 => if end_inv = '1' then current_state <= 0; end if;
      end case;
    end if;
  end process control_unit;

end rtl;
