----------------------------------------------------------------------------
-- Binary (square and multiply) exponentiation (exponentiation_montgomery.vhd)
--
-- LSB first
-- 
-- Computes the x^e mod f in GF(2**m)
-- Implements a sequential cincuit instantiang a multiplier and a squarer
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package exponentiation_montgomery_adv_package is
  constant M: integer := 233;
  constant N: integer := 233;
  constant COUNT_SQ: integer := 4; --number of clk cycles for squaring
  --constant F: std_logic_vector(M-1 downto 0):= "00011011"; --for M=8 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"001B"; --for M=16 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0101001B"; --for M=32 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"010100000101001B"; --for M=64 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0000000000000000010100000101001B"; --for M=128 bits
  --constant F: std_logic_vector(M-1 downto 0):= "000"&x"00000000000000000000000000000000000000C9"; --for M=163
  constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
  type matrix_reductionR is array (0 to M-1) of STD_LOGIC_VECTOR(M-2 downto 0);
  function r_by_r_mod_f return std_logic_vector;
end exponentiation_montgomery_adv_package;


package body exponentiation_montgomery_adv_package is
  function r_by_r_mod_f return std_logic_vector is
  variable R: matrix_reductionR;
  variable D: std_logic_vector(2*M-2 downto 0);
  variable C: std_logic_vector(M-1 downto 0);
begin
  for j in 0 to M-1 loop
     for i in 0 to M-2 loop
        R(j)(i) := '0'; 
     end loop;
  end loop;
  for j in 0 to M-1 loop
     R(j)(0) := f(j);
  end loop;
  for i in 1 to M-2 loop
     for j in 0 to M-1 loop
        if j = 0 then 
           R(j)(i) := R(M-1)(i-1) and R(j)(0);
        else
           R(j)(i) := R(j-1)(i-1) xor (R(M-1)(i-1) and R(j)(0)); 
        end if;
     end loop;
  end loop;

  D(0) := F(0); c(0) := F(0);
  for i in 1 to m-1 loop D(2*i-1) := '0'; D(2*i) := F(i); C(i) := D(i); end loop;
   
  for j in 0 to M-1 loop
    for i in 0 to M-2 loop
       C(j) := (C(j) xor (R(j)(i) and D(m+i)));
    end loop;
  end loop;
  return C;
end r_by_r_mod_f;

end  exponentiation_montgomery_adv_package;


-----------------------------------
-- Bynary exponentiation
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.exponentiation_montgomery_adv_package.all;
entity exponentiation_montgomery_adv is
port (
  A: in std_logic_vector (M-1 downto 0);
  E: in std_logic_vector (N-1 downto 0);
  clk, reset, start: in std_logic; 
  B: out std_logic_vector (M-1 downto 0);
  done: out std_logic
);
end exponentiation_montgomery_adv;

architecture rtl of exponentiation_montgomery_adv is
  component montgomery_comb_squarer is
  port (
    a: in std_logic_vector (M-1 downto 0);
    c: out std_logic_vector (M-1 downto 0)
  );
  end component montgomery_comb_squarer;

  component montgomery_mult is
  port (
    a, b: in std_logic_vector (M-1 downto 0);
    clk, reset, start: in std_logic; 
    z: out std_logic_vector (M-1 downto 0);
    done: out std_logic
  );
  end component montgomery_mult;

  signal inic, shift_r, first, last: std_logic;
  signal count: natural range 0 to N;
  signal count_sq: natural range 0 to COUNT_SQ;
  type states is range 0 to 12;
  signal current_state: states;
  signal cc, bb, new_b, new_c: std_logic_vector (M-1 downto 0);
  signal ee: std_logic_vector (N-1 downto 0);
  signal start_mult, start_sq, done_mult, done_sq, capt: std_logic;
  constant r_by_r: std_logic_vector (M-1 downto 0):= r_by_r_mod_f;

begin
  inst_mult: montgomery_mult port map (a => cc, b => bb,
     clk => clk, reset => reset, start => start_mult, Z => new_B, done => done_mult);

  inst_square: montgomery_comb_squarer port map (a => cc, c => new_c);

  counter_sq: process(reset, clk)
  begin
    if reset = '1' then count_sq <= 0; done_sq <= '0'; 
    elsif clk' event and clk = '1' then
      if start_sq = '1' then 
        count_sq <= 0;
      elsif count_sq = COUNT_SQ then
        done_sq <= '1';
      else
        count_sq <= count_sq + 1; 
      end if;
    end if;
  end process counter_sq;

  counter: process(reset, clk)
  begin
    if reset = '1' then count <= 0;
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        count <= 0;
      elsif shift_r = '1' then
        count <= count+1; 
    end if;
    end if;
  end process counter;

  sh_reg_e: process(reset, clk)
  begin
    if reset = '1' then ee <= (others => '0');
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        ee <= e;
      elsif shift_r = '1' then
        ee <= '0' & ee(N-1 downto 1); 
    end if;
    end if;
  end process sh_reg_e;

  register_c: process(reset, clk)
  begin
    if reset = '1' then cc <= (others => '0');
    elsif clk' event and clk = '1' then
      if first = '1' then 
        cc <= r_by_r;
      elsif inic = '1' then 
        cc <= new_b;
      elsif shift_r = '1' then
        cc <= new_c; 
      elsif last = '1' then
        cc <= (0 => '1', others => '0');
      end if;
    end if;
  end process register_c;

  register_b: process(reset, clk)
  begin
    if reset = '1' then bb <= (others => '0');
    elsif clk' event and clk = '1' then
      if first = '1' then 
        bb <= a;
      elsif inic = '1' then 
        bb <= F;
      elsif (shift_r = '1' and ee(0) = '1') or (capt = '1') then
        bb <= new_b; 
      end if;
    end if;
  end process register_b;

  b <= bb;

  control_unit: process(clk, reset, current_state, ee(0))
  begin
    case current_state is
      when 0 to 1 => inic <= '0'; shift_r <= '0'; done <= '1'; first <= '0'; start_sq <= '0'; start_mult <= '0'; last <= '0'; capt <= '0';
      when 2 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '1'; start_sq <= '0'; start_mult <= '0';last <= '0'; capt <= '0';
      when 3 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '1'; last <= '0'; capt <= '0';--for C=A.R
      when 4 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0';last <= '0'; capt <= '0';
      when 5 => inic <= '1'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0';last <= '0'; capt <= '0';
      when 6 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '1'; start_mult <= ee(0);last <= '0'; capt <= '0';
      when 7 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0'; last <= '0'; capt <= '0';
      when 8 => inic <= '0'; shift_r <= '1'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0'; last <= '0'; capt <= '0';
      when 9 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0'; last <= '1'; capt <= '0';
      when 10 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '1'; last <= '0'; capt <= '0';
      when 11 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0'; last <= '0'; capt <= '0';
      when 12 => inic <= '0'; shift_r <= '0'; done <= '0'; first <= '0'; start_sq <= '0'; start_mult <= '0'; last <= '0'; capt <= '1';
    end case;

    if reset = '1' then current_state <= 0;
    elsif clk'event and clk = '1' then
      case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3; --capture operands
      when 3 => current_state <= 4; --computes a.r
      when 4 => if (done_mult = '1') then current_state <= 5; end if; --whait finish C=A.R
      when 5 => current_state <= 6; --Capture C=A.R and B=F
      when 6 => current_state <= 7; --start operations
      when 7 => if (done_sq = '1' and (ee(0) = '0' or done_mult = '1')) then current_state <= 8; end if;
      when 8 => if count = N-1 then current_state <= 9; else current_state <= 6; end if;
      when 9 => current_state <= 10; --Adjust B= B/R
      when 10 => current_state <= 11; 
      when 11 => current_state <= 12;
      when 12 => if (done_mult = '1') then current_state <= 0; end if; --End
      end case;
    end if;
  end process control_unit;

end rtl;
