----------------------------------------------------------------------------
-- Binary (square and multiply) exponentiation (exponentiation_sq_mult.vhd)
--
-- LSB first
-- 
-- Computes the x^e mod f in GF(2**m)
-- Implements a sequential cincuit instantiang a multiplier and a squarer (combinational)
-- 
--
----------------------------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
package exponentiation_sq_mult_package is
  constant M: integer := 8;
  constant N: integer := 8;
  constant COUNT_SQ: integer := 2; --number of clk cycles for squaring (depends on F)
  constant F: std_logic_vector(M-1 downto 0):= "00011011"; --for M=8 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"001B"; --for M=16 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0101001B"; --for M=32 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"010100000101001B"; --for M=64 bits
  --constant F: std_logic_vector(M-1 downto 0):= x"0000000000000000010100000101001B"; --for M=128 bits
  --constant F: std_logic_vector(M-1 downto 0):= "000"&x"00000000000000000000000000000000000000C9"; --for M=163
  --constant F: std_logic_vector(M-1 downto 0):= (0=> '1', 74 => '1', others => '0'); --for M=233
end exponentiation_sq_mult_package;


-----------------------------------
-- Bynary exponentiation
-----------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use work.exponentiation_sq_mult_package.all;
entity exponentiation_sq_mult is
port (
  A: in std_logic_vector (M-1 downto 0);
  E: in std_logic_vector (N-1 downto 0);
  clk, reset, start: in std_logic; 
  B: out std_logic_vector (M-1 downto 0);
  done: out std_logic
);
end exponentiation_sq_mult;

architecture rtl of exponentiation_sq_mult is
  component interleaved_mult is
  port (
    A, B: in std_logic_vector (M-1 downto 0);
    clk, reset, start: in std_logic; 
    Z: out std_logic_vector (M-1 downto 0);
    done: out std_logic
  );
  end component interleaved_mult;

  component classic_squarer  is
  port (
    a: in std_logic_vector (M-1 downto 0);
    c: out std_logic_vector (M-1 downto 0)
  );
  end component classic_squarer; 

  signal inic, shift_r, ce_c: std_logic;
  signal count: natural range 0 to N;
  signal count_sq: natural range 0 to COUNT_SQ;
  type states is range 0 to 5;
  signal current_state: states;
  signal cc, bb, new_b, new_c: std_logic_vector (M-1 downto 0);
  signal ee: std_logic_vector (N-1 downto 0);
  signal start_mult, start_sq, done_mult, done_sq: std_logic;

begin

  inst_mult: interleaved_mult port map (A => cc, B => bb,
     clk => clk, reset => reset, start => start_mult, Z => new_B, done => done_mult);

  inst_square: classic_squarer port map (a => cc, c => new_c);

  counter_sq: process(reset, clk)
  begin
    if reset = '1' then count_sq <= 0; done_sq <= '0'; 
    elsif clk' event and clk = '1' then
      if start_sq = '1' then 
        count_sq <= 0;
      elsif count_sq = COUNT_SQ then
        done_sq <= '1';
      else
        count_sq <= count_sq + 1; 
      end if;
    end if;
  end process counter_sq;

  counter: process(reset, clk)
  begin
    if reset = '1' then count <= 0;
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        count <= 0;
      elsif shift_r = '1' then
        count <= count+1; 
    end if;
    end if;
  end process counter;

  sh_reg_e: process(reset, clk)
  begin
    if reset = '1' then ee <= (others => '0');
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        ee <= e;
      elsif shift_r = '1' then
        ee <= '0' & ee(N-1 downto 1); 
    end if;
    end if;
  end process sh_reg_e;

  register_c: process(reset, clk)
  begin
    if reset = '1' then cc <= (others => '0');
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        cc <= a;
      elsif shift_r = '1' then
        cc <= new_c; 
      end if;
    end if;
  end process register_c;

  register_b: process(reset, clk)
  begin
    if reset = '1' then bb <= (0 => '1', others => '0');
    elsif clk' event and clk = '1' then
      if inic = '1' then 
        bb <= (0 => '1', others => '0');
      elsif shift_r = '1' and ee(0) = '1' then
        bb <= new_b; 
      end if;
    end if;
  end process register_b;

  b <= bb;

  control_unit: process(clk, reset, current_state, ee(0))
  begin
    case current_state is
      when 0 to 1 => inic <= '0'; shift_r <= '0'; done <= '1'; ce_c <= '0'; start_sq <= '0'; start_mult <= '0';
      when 2 => inic <= '1'; shift_r <= '0'; done <= '0'; ce_c <= '0'; start_sq <= '0'; start_mult <= '0';
      when 3 => inic <= '0'; shift_r <= '0'; done <= '0'; ce_c <= '1'; start_sq <= '1'; start_mult <= ee(0);
      when 4 => inic <= '0'; shift_r <= '0'; done <= '0'; ce_c <= '1'; start_sq <= '0'; start_mult <= '0';
      when 5 => inic <= '0'; shift_r <= '1'; done <= '0'; ce_c <= '1'; start_sq <= '0'; start_mult <= '0';
    end case;

    if reset = '1' then current_state <= 0;
    elsif clk'event and clk = '1' then
      case current_state is
      when 0 => if start = '0' then current_state <= 1; end if;
      when 1 => if start = '1' then current_state <= 2; end if;
      when 2 => current_state <= 3; --capture operands
      when 3 => current_state <= 4; --start operations
      when 4 => if (done_sq = '1' and (ee(0) = '0' or done_mult = '1')) then current_state <= 5; end if;
      when 5 => if count = N-1 then current_state <= 0; else current_state <= 3; end if;
      end case;
    end if;
  end process control_unit;

end rtl;
